#include <stdio.h>
#include <stdlib.h>

#include <inttypes.h>

#include <string.h>

#include <assert.h>

#include <err.h>

#include <sys/queue.h>
#include <netdb.h>

#include <event.h>

#include "socket.h"
#include "bufio.h"
#include "bufio-socket.h"
#include "lookup.h"
#include "tls.h"

#define USE_TLS	1



static void main_gets_catch(struct bufio *io, void *line, size_t len, enum bufio_errno err, void *fd) {
	static char buf[1024];
	int n;

	if (len > 0) {
		for (n = len - 1; n > 0; n--)
			if (((char *)line)[n] != '\r' && ((char *)line)[n] != '\n')
				break;

		warnx("read %d bytes: %.*s", (int)len, n + 1, line);
	}

	if (err != 0) {
		if (BUFIO_EOF(err)) {
			bufio_close(io);

			warnx("got EOF");
		
			return /* void */;
		} else
			errx(EXIT_FAILURE, "bufio_gets: %s", bufio_strerror(io));
	}

	bufio_gets(io, buf, sizeof buf, &main_gets_catch, io, &(struct timeval){ 2, 0 });

	return /* void */;
} /* main_gets_catch() */


static void main_write_catch(struct bufio *io, void *request, size_t len, enum bufio_errno err, void *fd) {
	static char buf[1024];

	if (err != 0)
		errx(EXIT_FAILURE, "bufio_write: %s", bufio_strerror(io));

	warnx("wrote %d bytes", (int)len);

	if (len != strlen(request)) {
		bufio_write(io, (char *)request + len, strlen((char *)request + len), &main_write_catch, io, &(struct timeval){ 2, 0 });

		return /* void */;
	}

	bufio_gets(io, buf, sizeof buf, &main_gets_catch, io, &(struct timeval){ 2, 0 });

	return /* void */;
} /* main_write_catch() */


static void main_secure_catch(struct tls *tls, enum tls_errno err, void *fd) {
	static const char request[]	= "GET / HTTP/1.0\n\n";
	static struct bufio_socket so;
	struct bufio *io;
	enum bufio_errno bio_errno;

	assert(err == TLS_ESUCCESS);

	warnx("caught TLS handshake with google.com");

	assert(bufio_socket_init(&so, (intptr_t)fd, 0, 0, 0, &bio_errno));

	bufio_socket_starttls(&so, tls);

	assert(io = bufio_open(0, 0, &bio_errno));

	bufio_set_source(io, bufio_socket_to_source(&so));
	bufio_set_sink(io, bufio_socket_to_sink(&so));

	bufio_write(io, request, sizeof request - 1, &main_write_catch, io, &(struct timeval){ 2, 0 });

	return /* void */;
} /* main_secure_catch() */


static void main_connect_catch(int fd, struct connect_info *i, enum socket_errno err, void *arg) {
#if USE_TLS
	static const char request[]	= "GET / HTTP/1.0\n\n";
	static struct tls_context *tlsctx;
	struct tls *tls;

	assert(err == SOCKET_ESUCCESS);

	warnx("caught connect to google.com");

	socket_close(arg);

	assert(tlsctx = tls_context_open(0, 0, 0));
	assert(tls = tls_open(fd, tlsctx, 0, 0, 0));

	tls_connect(tls, &main_secure_catch, (void *)(intptr_t)fd, 0);

	return /* void */;
#else
	static const char request[]	= "GET / HTTP/1.0\n\n";
	static struct bufio_socket so;
	struct bufio *io;
	enum bufio_errno bio_errno;

	assert(err == TLS_ESUCCESS);

	warnx("caught TLS handshake with google.com");

	assert(bufio_socket_init(&so, (intptr_t)fd, 0, 0, 0, &bio_errno));

	assert(io = bufio_open(0, 0, &bio_errno));

	bufio_set_source(io, bufio_socket_to_source(&so));
	bufio_set_sink(io, bufio_socket_to_sink(&so));

	bufio_write(io, request, sizeof request - 1, &main_write_catch, io, &(struct timeval){ 2, 0 });

	return /* void */;
#endif
} /* main_connect_catch() */


int main(void) {
	struct socket *s;
	struct socket_name soname;

	event_init();
	lookup_init(0, 0, 0);
	tls_init();

	if (0 == (s = socket_open(0, 0, 0)))
		err(EXIT_FAILURE, "socket_open");

#if USE_TLS
	soname_set_host(&soname, "google.com", "https");
#else
	soname_set_host(&soname, "google.com", "http");
#endif

	socket_connect(s, &soname, &main_connect_catch, s, 0);

	event_loop(0);

	warnx("exiting");

	return 0;
} /* main() */

